import json

from . import Loader, I18nFileLoadError


class JsonLoader(Loader):
    """class to load json files"""
    def __init__(self):
        super(JsonLoader, self).__init__()

    def parse_file(self, file_content: str) -> dict:
        try:
            return json.loads(file_content)
        except json.JSONDecodeError as e:
            raise I18nFileLoadError("invalid JSON: {0}".format(e.args[0])) from e



__all__ = ("add", "get", "has", "clear")

from typing import Optional, Union, Tuple, Dict

from . import config

TranslationType = Union[str, Dict[str, str], Tuple[str, ...], Tuple[Dict[str, str], ...]]
container: Dict[str, Dict[str, TranslationType]] = {}


def add(
    key: str,
    value: TranslationType,
    locale: Optional[str] = None,
) -> None:
    """
    Adds translation to cache

    :param key: Translation key
    :param value: Translation
    :param locale: Locale (optional). Uses default if not provided
    """

    if locale is None:
        locale = config.get('locale')
    container.setdefault(locale, {})[key] = value


def has(key: str, locale: Optional[str] = None) -> bool:
    if locale is None:
        locale = config.get('locale')
    return key in container.get(locale, {})


def get(key: str, locale: Optional[str] = None) -> TranslationType:
    if locale is None:
        locale = config.get('locale')
    return container[locale][key]


def clear(locale: Optional[str] = None) -> None:
    if locale is None:
        container.clear()
    elif locale in container:
        container[locale].clear()



__all__: tuple = ("Loader", "PythonLoader", "I18nFileLoadError")

from .loader import Loader
from ..errors import I18nFileLoadError
from .python_loader import PythonLoader
from .. import config
if config.json_available:
    from .json_loader import JsonLoader
    __all__ += ("JsonLoader",)
if config.yaml_available:
    from .yaml_loader import YamlLoader
    __all__ += ("YamlLoader",)

del config



__all__ = ("TranslationFormatter", "StaticFormatter", "FilenameFormat", "expand_static_refs")

from re import compile, escape
try:
    from re import Match
except ImportError:
    # Python 3.6 doesn't have this
    Match = type(compile("").match(""))  # type: ignore
from string import Template, Formatter as _Fmt
from typing import Any, Iterable, Optional, Set, Callable, Tuple, TypeVar
from collections.abc import Mapping

from . import config, translations
from .translations import TranslationType
from .translator import pluralize
from .errors import I18nInvalidStaticRef, I18nInvalidFormat
from .custom_functions import get_function


class Formatter(
    Template,
    Mapping,
    metaclass=type(  # type: ignore[misc]
        "FormatterMeta",
        tuple(c for c in map(type, (Template, Mapping)) if c != type),
        {},
    ),
):
    delimiter = config.get("placeholder_delimiter")

    def __init__(self, translation_key: str, locale: str, value: TranslationType, kwargs: dict):
        super().__init__(value)  # type: ignore[arg-type]
        self.translation_key = translation_key
        self.locale = locale
        self.kwargs = kwargs

    def substitute(self) -> str:  # type: ignore[override]
        return super().substitute(self)

    def safe_substitute(self) -> str:  # type: ignore[override]
        return super().safe_substitute(self)

    def _format_str(self) -> str:
        raise NotImplementedError

    def format(self) -> TranslationType:
        if isinstance(self.template, str):
            return self._format_str()
        if isinstance(self.template, dict):
            result = {}
            for k, v in self.template.items():
                self.template = v
                result[k] = self.format()
            return result
        # assuming list/tuple
        result = []
        for v in self.template:
            self.template = v
            result.append(self.format())
        return tuple(result)

    def __getitem__(self, key: str) -> Any:
        return self.kwargs[key]

    def __len__(self):
        return self.kwargs.__len__()

    def __iter__(self):
        return self.kwargs.__iter__()


class TranslationFormatter(Formatter):
    idpattern = r"""
        \w+                      # name
        (
            \(
                [^\(\){}]*       # arguments
            \)
        )?
    """

    def __init__(self, translation_key: str, locale: str, value: TranslationType, kwargs: dict):
        super().__init__(translation_key, locale, value, kwargs)
        self.pluralized = False

    def format(self) -> TranslationType:
        if not self.pluralized and "count" in self.kwargs:
            if isinstance(self.template, tuple):
                self.template = tuple(
                    pluralize(
                        self.translation_key,
                        self.locale,
                        i,
                        self.kwargs["count"],
                    )
                    for i in self.template
                )
            else:
                self.template = pluralize(
                    self.translation_key,
                    self.locale,
                    self.template,
                    self.kwargs["count"],
                )
            self.pluralized = True
        return super().format()

    def _format_str(self) -> str:
        if config.get("on_missing_placeholder"):
            return self.substitute()
        else:
            return self.safe_substitute()

    def __getitem__(self, key: str) -> Any:
        try:
            name, _, args = key.partition("(")
            if args:
                f = get_function(name, self.locale)
                if f:
                    i = f(**self.kwargs)
                    arg_list = args.strip(")").split(config.get("argument_delimiter"))
                    try:
                        return arg_list[i]
                    except (IndexError, TypeError) as e:
                        raise ValueError(
                            "No argument {0!r} for function {1!r} (in {2!r})".format(
                                i, name, self.template
                            )
                        ) from e
                raise KeyError(
                    "No function {0!r} found for locale {1!r} (in {2!r})".format(
                        name, self.locale, self.template
                    )
                )
            return super().__getitem__(key)
        except KeyError:
            on_missing = config.get("on_missing_placeholder")
            if not on_missing or on_missing == "error":
                raise
            return on_missing(self.translation_key, self.locale, self.template, key)


class StaticFormatter(Formatter):
    idpattern = r"""
        ({})
        |
        ({}\w+)+
    """.format(
        TranslationFormatter.idpattern,
        escape(config.get("namespace_delimiter")),
    )

    def __init__(self, translation_key: str, locale: str, value: TranslationType):
        super().__init__(translation_key, locale, value, {})
        self.path = translation_key.split(config.get("namespace_delimiter"))

    def _format_str(self) -> str:
        return self.safe_substitute()

    def __getitem__(self, key: str) -> Any:
        delim = config.get("namespace_delimiter")
        full_key = key.lstrip(delim)
        if full_key == key:
            # not a static reference, skip
            raise KeyError()

        for i in range(1, len(self.path) + 1):
            try:
                # keep expanding in case of nested references
                # python will throw an exception if there's a recursive reference
                expand_static_refs((full_key,), self.locale)
                return translations.get(full_key, self.locale)
            except KeyError:
                full_key = delim.join(self.path[:i]) + key

        # try to search in other files
        from .resource_loader import search_translation

        full_key = key.lstrip(delim)
        search_translation(full_key, self.locale)
        try:
            return translations.get(full_key, self.locale)
        except KeyError:
            raise I18nInvalidStaticRef(
                "no value found for static reference {!r} (in {!r})"
                .format(key, self.translation_key),
            )


def expand_static_refs(keys: Iterable[str], locale: str) -> None:
    for key in keys:
        tr = translations.get(key, locale)
        tr = StaticFormatter(key, locale, tr).format()
        translations.add(key, tr, locale)


# This is (hopefully) a temporary workaround
# https://github.com/python/mypy/issues/15848
StrOrLiteralStr = TypeVar("StrOrLiteralStr", str, str)


class FilenameFormat(_Fmt):
    def __init__(self, template: str, variables: dict):
        super().__init__()
        self.template = template
        self.variables = variables
        self.used_variables: Set[str] = set()
        self.pattern = compile(super().format(template))

    @property
    def format(self) -> Callable[..., str]:
        return self.template.format

    @property
    def match(self) -> Callable[[str], Optional[Match]]:
        return self.pattern.fullmatch

    def __getattr__(self, name: str) -> bool:
        if name.startswith("has_"):
            _, _, var_name = name.partition("_")
            if var_name in self.variables:
                return var_name in self.used_variables
        raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}")

    def parse(self, s: StrOrLiteralStr) -> Iterable[Tuple[StrOrLiteralStr, None, None, None]]:
        for text, field, spec, conversion in super().parse(s):
            if spec or conversion:
                raise I18nInvalidFormat("Can't apply format spec or conversion in filename format")
            text = escape(text)
            if field is not None:
                try:
                    text += f"(?P<{field}>{self.variables[field]})"
                except KeyError as e:
                    raise I18nInvalidFormat(f"Unknown placeholder in filename format: {e}") from e
                self.used_variables.add(field)
            yield text, None, None, None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.template!r}, {self.variables!r})"



import os.path
from importlib import util

from . import Loader, I18nFileLoadError


class PythonLoader(Loader):
    """class to load python files"""

    def __init__(self):
        super(PythonLoader, self).__init__()

    def load_file(self, filename: str) -> dict:  # type: ignore[override]
        _, name = os.path.split(filename)
        module_name, _ = os.path.splitext(name)
        try:
            spec = util.spec_from_file_location(module_name, filename)
            # loader is guaranteed to be not None if spec is not None
            # but we check it anyway to make mypy happy
            assert spec is not None and spec.loader, "spec is None"
            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return vars(module)
        except Exception as e:
            raise I18nFileLoadError("error loading file {!r}: {}".format(filename, e)) from e

    def parse_file(self, file_content: dict) -> dict:  # type: ignore[override]
        return file_content



import io
import os.path
from typing import Optional, Dict

from .. import config
from ..errors import I18nFileLoadError


class Loader(object):
    """Base class to load resources"""

    loaded_files: Dict[str, Optional[dict]] = {}

    def __init__(self):
        super(Loader, self).__init__()

    def load_file(self, filename: str) -> str:
        """
        Reads content from file

        :param filename: The file to read
        :return: Content of the file
        :raises I18nFileLoadError: If loading wasn't successful
        """

        try:
            with io.open(filename, 'r', encoding=config.get('encoding')) as f:
                return f.read()
        except IOError as e:
            raise I18nFileLoadError(
                "error loading file {0}: {1}".format(filename, e.strerror),
            ) from e

    def parse_file(self, file_content: str) -> dict:
        """
        Parses file content to dict. Must be implemented in subclasses

        :param file_content: Content of the file
        :return: Parsed content
        :raises I18nFileLoadError: If parsing wasn't successful
        """

        raise NotImplementedError(
            "the method parse_file has not been implemented for class {0}".format(
                self.__class__.__name__,
            ),
        )

    def check_data(self, data: dict, root_data: Optional[str]) -> bool:
        """
        Checks if `root_data` is present in the content

        :param data: Parsed content of the file
        :param root_data: Data element to be checked. If `None`, check always succeeds
        :return: `True` if `root_data` is present, `False` otherwise
        """

        return True if root_data is None else root_data in data

    def get_data(self, data: dict, root_data: Optional[str]) -> dict:
        """
        Extracts `root_data` from `data`

        :param data: Full parsed data
        :param root_data: Part of the data to extract
        :return: Extracted data
        """

        # use .pop to remove used data from cache
        return data if root_data is None else data.pop(root_data)

    def load_resource(
        self,
        filename: str,
        root_data: Optional[str],
        remember_content: bool,
    ) -> dict:
        """
        Main function for resource loading

        Manages caching

        :param filename: File to load
        :param root_data: Part of data to extract
        :param remember_content: Whether to save other parts of data in cache
        :return: Loaded data
        :raises I18nFileLoadError: If loading wasn't successful
        """

        filename = os.path.abspath(filename)
        if filename in self.loaded_files:
            data = self.loaded_files[filename]
            if not data:
                # cache is missing or exhausted
                return {}
        else:
            file_content = self.load_file(filename)
            data = self.parse_file(file_content)
        if not self.check_data(data, root_data):
            raise I18nFileLoadError(
                "error getting data from {0}: {1} not defined".format(filename, root_data),
            )
        enable_memoization = config.get('enable_memoization')
        if enable_memoization:
            if remember_content:
                self.loaded_files[filename] = data
            else:
                self.loaded_files[filename] = None
        return self.get_data(data, root_data)



__all__ = (
    "Loader",
    "register_loader",
    "init_loaders",
    "load_config",
    "load_everything",
    "unload_everything",
    "reload_everything",
    "search_translation",
)

import os.path
from typing import Type, Iterable, Optional, List, Set, Union

from . import config
from .loaders import Loader, I18nFileLoadError
from .errors import I18nLockedError
from . import translations, formatters

loaders = {}

PLURALS = {"zero", "one", "few", "many"}


def register_loader(loader_class: Type[Loader], supported_extensions: Iterable[str]) -> None:
    """
    Registers loader for files

    :param loader_class: Loader to register
    :param supported_extensions: Iterable of file extensions that the loader supports
    :raises ValueError: If `loader_class` is not a subclass of `i18n.Loader`
    """

    if not issubclass(loader_class, Loader):
        raise ValueError("loader class should be subclass of i18n.Loader")

    loader = loader_class()
    for extension in supported_extensions:
        loaders[extension] = loader


def load_resource(filename: str, root_data: Optional[str], remember_content: bool = False) -> dict:
    extension = os.path.splitext(filename)[1][1:]
    if extension not in loaders:
        raise I18nFileLoadError("no loader available for extension {0}".format(extension))
    return loaders[extension].load_resource(filename, root_data, remember_content)


def init_loaders():
    """Sets default loaders"""

    init_python_loader()
    if config.yaml_available:
        init_yaml_loader()
    if config.json_available:
        init_json_loader()


def init_python_loader():
    from .loaders import PythonLoader
    register_loader(PythonLoader, ["py"])


def init_yaml_loader():
    from .loaders import YamlLoader
    register_loader(YamlLoader, ["yml", "yaml"])


def init_json_loader():
    from .loaders import JsonLoader
    register_loader(JsonLoader, ["json"])


def load_config(filename: str) -> None:
    """
    Loads configuration from file

    :param filename: File containing configuration
    """

    settings_data = load_resource(filename, "settings")
    for key, value in settings_data.items():
        config.set(key, value)


def get_namespace_from_filepath(filename: str) -> str:
    namespace = os.path.dirname(filename).strip(os.sep).replace(
        os.sep,
        config.get("namespace_delimiter"),
    )
    format = config.get('filename_format')
    if format.has_namespace:
        filename_match = format.match(os.path.basename(filename))
        if namespace:
            namespace += config.get('namespace_delimiter')
        namespace += filename_match.group("namespace")
    return namespace


def load_translation_file(filename: str, base_directory: str, locale: Optional[str] = None) -> None:
    if locale is None:
        locale = config.get('locale')
    skip_locale_root_data = config.get('skip_locale_root_data')
    root_data = None if skip_locale_root_data else locale
    # if the file isn't dedicated to one locale and may contain other `root_data`s
    remember_content = not config.get("filename_format").has_locale and root_data
    translations_dic = load_resource(
        os.path.join(base_directory, filename),
        root_data,
        bool(remember_content),
    )
    namespace = get_namespace_from_filepath(filename)
    loaded = load_translation_dic(translations_dic, namespace, locale)
    formatters.expand_static_refs(loaded, locale)


_locked: Union[bool, Set[Union[str, None]]] = False


def _check_locked(locale: Optional[str]) -> bool:
    return _locked if isinstance(_locked, bool) else locale in _locked


def load_everything(locale: Optional[str] = None, *, lock: bool = False) -> None:
    """
    Loads all translations

    If locale is provided, loads translations only for that locale

    :param locale: Locale (optional)
    :param lock: Whether to lock translations after loading.
    Locking disables further searching for missing translations
    """

    global _locked

    if _check_locked(locale):
        raise I18nLockedError("Translations were locked, use unload_everything() to unlock")

    for directory in config.get("load_path"):
        recursive_load_everything(directory, "", locale)

    if not lock:
        return

    if locale:
        if isinstance(_locked, bool):
            _locked = {None, locale}
        else:
            _locked.add(locale)
    else:
        _locked = True


def unload_everything():
    """Clears all cached translations"""

    global _locked

    translations.clear()
    Loader.loaded_files.clear()
    _locked = False


def reload_everything(*, lock: bool = False) -> None:
    """
    Shortcut for `unload_everything()` + `load_everything()`

    :param lock: Passed to `load_everything()`, see its description for more information
    """

    unload_everything()
    load_everything(lock=lock)


def load_translation_dic(dic: dict, namespace: str, locale: str) -> Iterable[str]:
    loaded: List[str] = []
    if namespace:
        namespace += config.get('namespace_delimiter')
    for key, value in dic.items():
        full_key = namespace + key
        if isinstance(value, dict) and len(PLURALS.intersection(value)) < 2:
            loaded.extend(load_translation_dic(value, full_key, locale))
        else:
            translations.add(full_key, value, locale)
            loaded.append(full_key)
    return loaded


def search_translation(key: str, locale: Optional[str] = None) -> None:
    if locale is None:
        locale = config.get('locale')
    if _check_locked(locale):
        return
    splitted_key = key.split(config.get('namespace_delimiter'))
    namespace = splitted_key[:-1]
    for directory in config.get("load_path"):
        recursive_search_dir(namespace, "", directory, locale)


def recursive_search_dir(
    splitted_namespace: List[str],
    directory: str,
    root_dir: str,
    locale: str,
) -> None:
    namespace = splitted_namespace[0] if splitted_namespace else ""
    seeked_file = config.get("filename_format").format(
        namespace=namespace,
        locale=locale,
        format=config.get("file_format"),
    )
    dir_content = os.listdir(os.path.join(root_dir, directory))
    if seeked_file in dir_content:
        load_translation_file(os.path.join(directory, seeked_file), root_dir, locale)
    elif namespace in dir_content:
        recursive_search_dir(
            splitted_namespace[1:],
            os.path.join(directory, namespace),
            root_dir,
            locale,
        )


def recursive_load_everything(root_dir: str, directory: str, locale: Optional[str]) -> None:
    dir_ = os.path.join(root_dir, directory)
    for f in os.listdir(dir_):
        path = os.path.join(dir_, f)
        if os.path.isfile(path):
            if os.path.splitext(path)[1][1:] != config.get("file_format"):
                continue
            format_match = config.get("filename_format").match(f)
            if not format_match:
                continue
            requested_locale = locale
            file_locale = format_match.groupdict().get("locale", requested_locale)
            if requested_locale is None:
                requested_locale = file_locale
            if requested_locale is not None:
                if requested_locale == file_locale:
                    load_translation_file(
                        os.path.join(directory, f),
                        root_dir,
                        requested_locale,
                    )
            elif not config.get("skip_locale_root_data"):
                file_content = load_resource(path, None, False)
                for loc, dic in file_content.items():
                    if isinstance(dic, dict):
                        loaded = load_translation_dic(
                            dic,
                            get_namespace_from_filepath(os.path.join(directory, f)),
                            loc,
                        )
                        formatters.expand_static_refs(loaded, loc)
            else:
                raise I18nFileLoadError(
                    f"Cannot identify locales for {path!r}:"
                    " filename_format doesn't include locale"
                    " and skip_locale_root_data is set to True"
                )
        elif os.path.isdir(path):  # pragma: no branch
            recursive_load_everything(
                root_dir,
                os.path.join(directory, f),
                locale,
            )



from typing import Type, Union

import yaml
from yaml.loader import __all__ as _known_loaders  # type: ignore[attr-defined]

from . import Loader, I18nFileLoadError


class YamlLoader(Loader):
    """class to load yaml files"""

    loader: Union[  # type: ignore[valid-type]
        tuple(Type[getattr(yaml, i)] for i in _known_loaders)
    ] = yaml.BaseLoader

    def __init__(self):
        super(YamlLoader, self).__init__()

    def parse_file(self, file_content: str) -> dict:
        try:
            return yaml.load(file_content, Loader=self.loader)
        except yaml.YAMLError as e:
            raise I18nFileLoadError("invalid YAML: {0}".format(str(e))) from e



__all__ = ("set", "get")

from typing import Any
from importlib import reload as _reload

try:
    __import__("yaml")
    yaml_available = True
except ImportError:
    yaml_available = False

try:
    __import__("json")
    json_available = True
except ImportError:
    json_available = False

# try to get existing path object
# in case if config is being reloaded
try:
    from . import load_path
    load_path.clear()
except ImportError:
    load_path = []


FILENAME_VARS = dict.fromkeys(
    ("namespace", "locale", "format"),
    r"\w+",
)


settings = {
    'filename_format': '{namespace}.{locale}.{format}',
    'file_format': 'yml' if yaml_available else 'json' if json_available else 'py',
    'available_locales': ['en'],
    'load_path': load_path,
    'locale': 'en',
    'fallback': 'en',
    'placeholder_delimiter': '%',
    'on_missing_translation': None,
    'on_missing_placeholder': None,
    'on_missing_plural': None,
    'encoding': 'utf-8',
    'namespace_delimiter': '.',
    'plural_few': 5,
    'skip_locale_root_data': False,
    "enable_memoization": True,
    'argument_delimiter': '|'
}


def set(key: str, value: Any) -> None:
    """
    Sets config value

    :param key: Setting to set
    :param value: New value
    :raises KeyError: If `key` is not a valid key
    """

    if key not in settings:
        raise KeyError("Invalid setting: {0}".format(key))
    elif key == 'load_path':
        load_path.clear()
        load_path.extend(value)
        return
    elif key == 'filename_format':
        from .formatters import FilenameFormat

        value = FilenameFormat(value, FILENAME_VARS)

    settings[key] = value

    if key in ('placeholder_delimiter', 'namespace_delimiter'):
        from . import formatters

        _reload(formatters)


def get(key: str) -> Any:
    """
    Gets config value

    :param key: Setting to get
    :return: Associated value
    :raises KeyError: If `key` is not a valid key
    """

    return settings[key]


# initialize FilenameFormat
set('filename_format', get('filename_format'))



__all__ = ("t",)

from typing import Any, Dict, Union, Tuple, Optional, overload
try:
    from typing import SupportsIndex
except ImportError:
    SupportsIndex = int  # type: ignore

from . import config
from . import resource_loader
from . import translations, formatters


def t(
    key: str,
    *,
    locale: Optional[str] = None,
    **kwargs: Any,
) -> Union[str, "LazyTranslationTuple"]:
    """
    Main translation function

    Searches for translation in files if it's not already in cache
    Tries fallback locale if search fails and fallback is set
    If that also fails:
      - Returns original key if `on_missing_translation` is not set
      - Raises `KeyError` if it's set to `"error"`
      - Returns result of calling it if it's set to a function

    :param key: Translation key
    :param locale: Locale to translate to (optional)
    :param **kwargs: Keyword arguments used to interpolate placeholders
    (including `count` for pluralization)
    :return: The translation, return value of `on_missing_translation` or the original key
    :raises KeyError: If translation wasn't found and `on_missing_translation` is set to `"error"`
    """

    if not locale:
        locale = config.get("locale")
    try:
        return translate(key, locale=locale, **kwargs)  # type: ignore[arg-type]
    except KeyError:
        resource_loader.search_translation(key, locale)
        if translations.has(key, locale):
            return translate(key, locale=locale, **kwargs)  # type: ignore[arg-type]
        fallback = config.get("fallback")
        if fallback and fallback != locale:
            return t(key, locale=fallback, **kwargs)
    on_missing = config.get('on_missing_translation')
    if on_missing == "error":
        raise KeyError('key {0} not found'.format(key))
    elif on_missing:
        return on_missing(key, locale, **kwargs)
    else:
        return key


class LazyTranslationTuple(tuple):
    translation_key: str
    locale: str
    kwargs: dict

    def __new__(
        cls,
        translation_key: str,
        locale: str,
        value: tuple,
        kwargs: dict,
    ) -> "LazyTranslationTuple":
        obj = super().__new__(cls, value)
        obj.translation_key = translation_key
        obj.locale = locale
        obj.kwargs = kwargs
        return obj

    @overload
    def __getitem__(self, key: SupportsIndex) -> str: ...

    @overload
    def __getitem__(self, key: slice) -> Tuple[str, ...]: ...

    def __getitem__(self, key: Union[SupportsIndex, slice]) -> Union[str, Tuple[str, ...]]:
        return formatters.TranslationFormatter(  # type: ignore[return-value]
            self.translation_key,
            self.locale,
            super().__getitem__(key),
            self.kwargs,
        ).format()


def translate(key: str, locale: str, **kwargs: Any) -> Union[str, LazyTranslationTuple]:
    translation = translations.get(key, locale=locale)
    if isinstance(translation, tuple):
        return LazyTranslationTuple(key, locale, translation, kwargs)
    else:
        return formatters.TranslationFormatter(
            key, locale, translation, kwargs
        ).format()  # type: ignore[return-value]


def pluralize(key: str, locale: str, translation: Union[Dict[str, str], str], count: int) -> str:
    return_value = key
    try:
        if not isinstance(translation, dict):
            return_value = translation
            raise KeyError('use of count witouth dict for key {0}'.format(key))
        if count == 0:
            if 'zero' in translation:
                return translation['zero']
        elif count == 1:
            if 'one' in translation:
                return translation['one']
        elif count <= config.get('plural_few'):
            if 'few' in translation:
                return translation['few']
        if 'many' in translation:
            return translation['many']
        else:
            raise KeyError('"many" not defined for key {0}'.format(key))
    except KeyError:
        on_missing = config.get('on_missing_plural')
        if on_missing == "error":
            raise
        elif on_missing:
            return on_missing(key, locale, translation, count)
        else:
            return return_value



class I18nException(Exception):
    """Base class for i18n errors"""

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return self.value


class I18nFileLoadError(I18nException):
    """Raised when file load fails"""
    pass


class I18nInvalidStaticRef(I18nException):
    """Raised when static reference cannot be resolved"""
    pass


class I18nInvalidFormat(I18nException):
    """Raised when provided filename_format is invalid"""
    pass


class I18nLockedError(I18nException):
    """Raised when trying to load locked translations"""
    pass



import re
from urllib.parse import urljoin

from setuptools import setup


GITHUB_URL = "https://github.com/Krutyi-4el/i18nice"
long_description = open("README.md").read()
# links on PyPI should have absolute URLs
long_description = re.sub(
    r"(\[[^\]]+\]\()((?!https?:)[^\)]+)(\))",
    lambda m: m.group(1) + urljoin(GITHUB_URL, "blob/master/" + m.group(2)) + m.group(3),
    long_description,
)

setup(
    name='i18nice',
    version='0.9.0',
    description='Translation library for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Daniel Perez',
    author_email='tuvistavie@gmail.com',
    maintainer="Krutyi 4el",
    url=GITHUB_URL,
    download_url=urljoin(GITHUB_URL, "archive/master.zip"),
    license='MIT',
    packages=['i18n', 'i18n.loaders'],
    zip_safe=True,
    test_suite='i18n.tests',
    extras_require={
        'YAML': ["pyyaml>=3.10"],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Internationalization',
    ],
)



__all__ = ("add_function", "get_function")

from collections import defaultdict
from typing import Optional, Callable, Dict


Function = Callable[..., int]
global_functions: Dict[str, Function] = {}
locales_functions: Dict[str, Dict[str, Function]] = defaultdict(dict)


def add_function(name: str, func: Function, locale: Optional[str] = None) -> None:
    """
    Adds your function to placeholder functions

    Registers the function to locale functions if locale is given
    or to global (locale-independent) functions otherwise
    The function must accept all kwargs passed to `t` and return an int
    (index that will determine which placeholder argument will be used)

    :param name: Name used to register the function
    :param func: The function to register
    :param locale: Locale to which function will be bound (optional)
    """

    if locale:
        locales_functions[locale][name] = func
    else:
        global_functions[name] = func


def get_function(name: str, locale: Optional[str] = None) -> Optional[Function]:
    if locale and name in locales_functions[locale]:
        return locales_functions[locale][name]
    return global_functions.get(name)



__all__ = (
    "Loader",
    "register_loader",
    "init_default_loaders",
    "load_config",
    "load_everything",
    "unload_everything",
    "reload_everything",

    "I18nException",
    "I18nFileLoadError",
    "I18nInvalidStaticRef",
    "I18nInvalidFormat",

    "t",
    "add_translation",
    "add_function",

    "set",
    "get",

    "load_path",
)

from typing import List

from .resource_loader import (
    Loader,
    register_loader,
    init_loaders as init_default_loaders,
    load_config,
    load_everything,
    unload_everything,
    reload_everything,
)
from .errors import (
    I18nException,
    I18nFileLoadError,
    I18nInvalidStaticRef,
    I18nInvalidFormat,
)
from .translator import t
from .translations import add as add_translation
from .custom_functions import add_function
from .config import set, get

init_default_loaders()

load_path: List[str] = get("load_path")

del List
